/* atomic.h */

/*
 * Copyright 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014 by
 *	University of New Hampshire Interoperability Laboratory
 *	121 Technology Drive, Suite 2
 *	Durham, NH 03824-4716
 *	phone: 1-603-862-0090
 *	fax: 1-603-862-4181
 *	e-mail: info@iol.unh.edu
 * All rights reserved.
 *
 * This file is part of the UNH EXS implementation of the
 * Consortium Specification: Extended Sockets API (ES-API) Issue 1.0,
 * Document Number C050 published by The Open Group, January 2005,
 * and copyright 2005 by The Open Group.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING.GPL in the main directory of this source tree, or the
 * BSD license, available from the file COPYING.BSD in the main directory
 * of this source tree.
 *
 * This material is based upon work supported by the National Science
 * Foundation under Grant No. OCI-1127228.
 *
 * Any opinions, findings, and conclusions or recommendations
 * expressed in this material are those of the authors and do not
 * necessarily reflect the views of the National Science Foundation.
 */

#ifndef ATOMIC_H
#define ATOMIC_H

#if !defined(HAVE_GCC_SYNC_ATOMICS)
unsigned int
atomic_add_reset(unsigned int increment, unsigned int limit,
						volatile unsigned int *result);

unsigned int
atomic_add(int amount, volatile unsigned int *result);

unsigned int
atomic_set_bits(unsigned int mask, volatile unsigned int *result);

unsigned int
atomic_clear_bits(unsigned int mask, volatile unsigned int *result);

unsigned int
atomic_swap_if_match(unsigned int swap_value, unsigned int match_value,
						volatile unsigned int *result);

void *
atomic_swap_if_match_ptr(void *swap_value, void *match_value,
						void *volatile *result);
#else

static inline unsigned int
atomic_add_reset(unsigned int increment, unsigned int limit,
						volatile unsigned int *result)
{
	unsigned int v_old, v_new, v_reset;

	do {
		v_old = *result;
		v_reset = v_new = v_old + increment;
		if (v_new >= limit) {
			v_reset = 0;
		}
	} while (!__sync_bool_compare_and_swap(result, v_old, v_reset));

	return v_new;
}

static inline unsigned int
atomic_add(int amount, volatile unsigned int *result)
{
	return (amount >= 0) ? __sync_add_and_fetch(result, amount)
					: __sync_sub_and_fetch(result, -amount);
}

static inline unsigned int
atomic_set_bits(unsigned int mask, volatile unsigned int *result)
{
	return __sync_fetch_and_or(result, mask);
}

static inline unsigned int
atomic_clear_bits(unsigned int mask, volatile unsigned int *result)
{
	return __sync_fetch_and_and(result, ~mask);
}

static inline unsigned int
atomic_swap_if_match(unsigned int swap_value, unsigned int match_value,
						volatile unsigned int *result)
{
	return __sync_val_compare_and_swap(result, match_value, swap_value);
}

static inline void *
atomic_swap_if_match_ptr(void *swap_value, void *match_value, void *volatile *result)
{
	return __sync_val_compare_and_swap(result, match_value, swap_value);
}

#endif	/* !defined(HAVE_GCC_SYNC_ATOMICS) */

#endif	/* #ifndef ATOMIC_H */
